/***************************************************************************
                          ipcalc.c  -  description
                             -------------------
    begin                : Thu Dec 23 2004
    copyright            : (C) 2004 by Hubert Vansteenhuyse
    email                : hvansteenhuyse@yahoo.com
 **************************************************************************

    This file is part of subnet.

    Subnet is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    Subnet is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Foobar; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
****************************************************************************/

#include "subnet.h"
char    ipbuf[32];

//returns class ABCDE of mask
char getipclass(long ip){
	if( (ip & 0x80000000) == 0) return 'A';				//first bits class A = 0
	if( (ip & 0xC0000000) == 0x80000000) return 'B';	//first bits class B = 10
	if( (ip & 0xE0000000) == 0xC0000000) return 'C';	//first bits class C = 110
	if( (ip & 0xF0000000) == 0xE0000000) return 'D';	//first bits class D = 1110
	return 'E';
}


//returns the amount of bits used in a mask
int calcbits(long mask){
    int b, bits = 0;
    for(b = 0; b < 32; b++){ if(mask & (long)pow(2, b)) bits++; }
    return bits;
}


//returns the ip notation (eg 127.0.0.1) from a long
char* longtoip(long ip){
    int w,x,y,z;
    w = (ip >> 24) & 0xFF;
    x = (ip >> 16) & 0xFF;
    y = (ip >> 8) & 0xFF;
    z = ip & 0xFF;
    sprintf(ipbuf, "%d.%d.%d.%d", w, x, y, z);
    return ipbuf;
}


//returns the long for an ip eg: 127.0.0.1
long iptolong(char *ip){
    char tmp[16];
	long ipa = 0;
   	int len = (int)strlen(ip), byte, bcnt = 4 * 8, from = 0, to = 0;
   	while(to < len){
   	    from = to;
		memset(tmp, 0, 16);
		while(to < len){
			if(ip[to] == '.') break;
			tmp[strlen(tmp)] = ip[to];
			to++;
		}
		bcnt -= 8;
		byte = atoi(tmp);
		ipa += (long)(byte << bcnt);
		to++;
    }
    return ipa;
}

//returns the netmask for a cidr mask (eg: 8 = 255.0.0.0 converted to its long value
long cidrtomask(int thecidrmask){
    int _thecidrmask = thecidrmask;
    long retval = 0;
    
    if(thecidrmask < 8 || thecidrmask > 30) return retval;
    while(thecidrmask){
        retval += (long) pow(2,thecidrmask);
        thecidrmask--;
    }
    return retval << (31 - _thecidrmask);
}

//returns a string like nnnnnnn.nnnnnnn.ssshhhhhhh,hhhhh... based on the mask
char binrepresentation[64];
char* getbitsofip(void){
	int nbits = calcbits(mask), cnt = 0, c, from = 0;
	memset(binrepresentation, 0, 64);
 	for(c = 0; c < 32; c++){
 	    sprintf(binrepresentation + strlen(binrepresentation), "%c", (nbits > 0 ? 'n' : 'h'));
 	    cnt++;
 	    if(cnt > 7 && c < 31){ sprintf(binrepresentation + strlen(binrepresentation), "."); cnt = 0; }
 	    nbits--;
    }
    
    nbits = getsubnetbits();
    cnt = 0;
	switch( (int)getipclass(ip) ){
	    case (int)'A':
	        from = 8;
	        cnt = 1;
	        break;
	        
	    case (int)'B':
	        from = 16;
	        cnt = 2;
	        break;
	        
	    case (int)'C':
	    	from = 24;
	        cnt = 3;
	        break;
    }    

    if(from > 0){
        for(c = 0; c < nbits; c++){ 
        	if(binrepresentation[from + c + cnt] == '.') cnt++;
          	binrepresentation[from + cnt + c] = 's';
        }    
    }    
 	return binrepresentation;
}

//returns the digit of an ip address. eg: digit 2 of 193.19.136.1 = 19
int getipdigit(long ipaddress, int digit){
    char *ips = longtoip(ipaddress), *ipe;
    int c;

    if(digit > 4 || digit < 1) return 0;
	for(c = 1; c < digit; c++){
		while(ips[0] != '.') ips++;
		ips++;
	}
	ipe = ips;
	while(ipe[0] != '.' && ipe[0] != 0) ipe++;
	ipe[0] = 0;
	return atoi(ips);
    
}

//returns the magic number (= 256 minus LAST NONZERO NETMASK BYTE)
int magicnumber(void){
    int     ipdigit = (calcbits(mask) / 8) + 1,
            magic = getipdigit(mask, ipdigit);
    return (magic == 0 ? 1 : 256 - magic);
}

//returns the private/public, apipa or loopback type
char* iptype(int ipaddress){
    if( getipdigit(ipaddress, 1) == 127) return "loopback";
    if( getipdigit(ipaddress, 1) == 10 ) return "private";
    if( getipdigit(ipaddress, 1) == 172 && getipdigit(ipaddress, 2) >= 16 && getipdigit(ipaddress, 2) <= 31) return "private";
    if( getipdigit(ipaddress, 1) == 192 && getipdigit(ipaddress, 2) == 168) return "private";
    if( getipdigit(ipaddress, 1) == 169 && getipdigit(ipaddress, 2) == 254) return "apipa";
    return "public";
}

//returns the amount of subnet bits used 
int getsubnetbits(void){
    int mbits;
    
   	mbits = calcbits(mask);
	switch( (int)getipclass(ip) ){
	    case (int)'A':
	        return  mbits - 8;
	        break;
	        
	    case (int)'B':
	        return  mbits - 16;
	        break;
	        
	    case (int)'C':
	        return  mbits - 24;
	        break;
    }    
    return -1;
}

//returns the amount of subnets
long getnets(void){ return (getsubnetbits() == 0 ? 1 : (long)pow(2, getsubnetbits()) - 2); }

//returns the amount of hosts
long gethosts(void){ 
    int net = (1 + (int)getipclass(ip) - (int)'A') * 8;
    if(net > 24) return -1;
    return (long)pow(2, 32 - net - getsubnetbits()) - 2;
}

/*checks the ip format and value. returns 1 for ok, 0 for not ok */
int isvalidip(char *theip){
    int x, pcnt = 0;
    long newip;
    char ipdig[32];
    	memset(ipdig, 0, 32);
    	for(x = 0; x <= (int)strlen(theip); x++){
    	    if(ISDIGIT(theip[x]) )
         		ipdig[strlen(ipdig)] = theip[x];
       		else{
				if(strlen(ipdig) == 0) return 0;
       		    if(theip[x] != '.' && theip[x] != 0) return 0;
       		    if(atoi(ipdig) >= 255 || atoi(ipdig) < 0) return 0;
				pcnt++;
				if(pcnt > 4) return 0;
       		    memset(ipdig, 0, 32);
            }
            
    	    if(strlen(ipdig) > 3) return 0;
        }
        newip = iptolong(theip);
        if(	newip == 0 || newip == -1 || getipdigit(newip, 1) == 0) return 0;
    return 1;
}

/*checks the subnetmask format and value. returns 1 for ok, 0 for not ok */
int isvalidmask(char *theip){
    long themask;
    int mbits, b = 31;
    	themask = iptolong(theip);
    	if(themask == 0 || themask == -1) return 0;
    	mbits = calcbits(themask);
    	if(mbits > (32 - 2) ) return 0;
    	
    	//test consecutive bits with value 1
   	    while(themask & (long)pow(2, b) && b >= 0) b--;
		//test consecutive bits with value 0
  	    while(!(themask & (long)pow(2, b)) && b >= 0) b--;
		//if all the bits were not tested then invalid mask. ok = 111110000000... error: 1111100001000...
    	if(b != -1) return 0;
    	
    	//test the mask for the current ip class 
    	switch( (int)getipclass(ip) ){
    	    case (int)'A':
    	        if(mbits == 8) return 1;
    	        if(mbits < (8 + 2)) return 0;
    	        break;
    	        
    	    case (int)'B':
    	        if(mbits == 16) return 1;
    	        if(mbits < (16 + 2)) return 0;
    	        break;
    	        
    	    case (int)'C':
       	        if(mbits == 24) return 1;
    	        if(mbits < (24 + 2)) return 0;
    	        break;
        }    
    return 1;
}

//calculates the network address of the specified network
long ipcalcnet(long network){
    int maskbits = calcbits(mask);                      //hoeveel bits worden er in de huidige mask gebruikt
    int subnetbits = getsubnetbits();                   //hoeveel bits voor het subnet
    int networkbits = maskbits - subnetbits;            //hoeveel netwerkbits voor deze ip classe (8, 16 of 24)
    int hostbits = 32 - (networkbits + subnetbits);     //hoeveel bits voor de hosts 

    long basenet = ip & cidrtomask(networkbits);        //bereken het netwerk volgens de huidige classe van het ip
    long subnet = (network - (subnetbits == 0 ? 1 : 0) ) << hostbits; //maak de bitmask voor het gevraagde netwerk

    return basenet | subnet;    //tel de netwerk bitmask bij het netwerk op 
}

//puts specified host bits on 1
long hosts_on(int bits){
    long retval = 0;
    while(bits){
        retval += 1 << (bits - 1);
        bits--;
    }
    return retval;
}

//calculates the broadcast address of the specified network
long ipcalcbcast(long network){
    int maskbits = calcbits(mask);
    int subnetbits = getsubnetbits();
    int networkbits = maskbits - subnetbits;
    int hostbits = 32 - (networkbits + subnetbits);
    return ipcalcnet(network) | hosts_on(hostbits); 
}

//sets mask for nearest amount of networks
void useamount_net(unsigned long nets){
    int 	maskbits = calcbits(mask),
    		subnetbits = getsubnetbits(),
    		networkbits = maskbits - subnetbits,	// nu kennen we de ip class 
    		maxsubnetbits = 30 - networkbits,
    		b = 2;									// minimum # subnetbits = 2 
    		
    	if(nets > 1){
    		//choose the nearest amount of subnet bits for the wanted amount of networks 
       	    while(b < maxsubnetbits){
       	        if( (unsigned long)pow(2, b) - 2 >= nets) break;
                b++;
            }    
        }else
          b = 0;
        mask = cidrtomask(networkbits + b);
}

//sets mask for nearest amount of hosts
void useamount_host(unsigned long hosts){
    int 	maskbits = calcbits(mask),
    		subnetbits = getsubnetbits(),
    		networkbits = maskbits - subnetbits,	// nu kennen we de ip class 
			maxhostbits = 30 - networkbits,
			b = 2;									// minimum # hostbits = 2 
		
   	    while(b <= maxhostbits){
   	        if( (unsigned long)pow(2, b) - 2 >= hosts) break;
            b++;
        }
        if(b > maxhostbits)
			mask = cidrtomask(networkbits);
		else
			mask = cidrtomask(networkbits + (32 - networkbits) - b);
}

//returns the subnet of an ip 
unsigned long findnet(long theip){
    int 	maskbits = calcbits(mask),
    		subnetbits = getsubnetbits(),
    		networkbits = maskbits - subnetbits,
    		hostbits = 32 - (networkbits + subnetbits);

    long 	basemask = cidrtomask(networkbits),
    		basenet = ip & basemask,
      		theSubNet = 0;
    
    if((theip & basemask) != basenet) return -1;

	
	//weg met de netwerkbits
	theSubNet = theip - basenet;

	//weg met de hostbits
    theSubNet = theSubNet >> hostbits;
    
    return theSubNet;
}












